#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <bios.h>
#include <memory.h>
#include <errno.h>

typedef	unsigned long	ulong;
typedef unsigned int	uint;
typedef unsigned short	ushort;
typedef unsigned char	uchar;

/* Variables declarations */
uint _far buf0[256];
uint _far buf1[256];
uint _far buf2[512];
uint diskerr, biosdrv;

int byte_even, byte_odd, byte_out ;
FILE *mbr_file, *cfmbr_file, *int13r_file;

int main(int argc , char *argv[]) {
struct diskinfo_t diskreq;

    if(argc < 3) {
	printf(" cfi V1.1\n");
	printf(" Usage: cfi [MB|MR|CI DRV] \n");
	printf(" MB  - backup MBR to <mbr.bin> file\n");
	printf(" MR  - restore MBR from <mbr.bin> file\n");
	printf(" CI  - install CF fix to MBR\n");
	printf(" DRV - drive number as BIOS int13h call (80, 81, 82, 83)\n");
	return 0;
    }

    sscanf(argv[2], "%x", &biosdrv);
    if ((biosdrv < 0x80) || (biosdrv > 0x83)) {
	printf("Invalid drive number!\n");
	exit(EINVAL);
     }
	diskreq.drive    = biosdrv & 0xff;
	diskreq.head     = 0x00;
	diskreq.track    = 0x00;
	diskreq.sector   = 0x01;
	diskreq.nsectors = 0x01;
	diskreq.buffer   = buf0;

/*-----------------------select mb cmd option----------------------------*/
    if (!strnicmp(argv[1], "mb", 2)) { 
	if (mbr_file = fopen("mbr.bin","rb")) {
	    printf("<mbr.bin> file already exist!\n");
	    fclose(mbr_file);
	    exit (EEXIST);
	}

	/* read original mbr to buf0 */
	printf("Read disk %s MBR to <mbr.bin> ... ", argv[2]);
	diskerr = _bios_disk(_DISK_READ, &diskreq);
	if (diskerr & 0xff00) {
	    printf("\nError read sector %02d, code 0x%04x\n",
		diskreq.sector, diskerr);
	    exit(EXIT_FAILURE);
	}

	if (!(mbr_file = fopen("mbr.bin","wb"))) {
	    printf("\nError creating <mbr.bin> file!\n");
	    exit(ENFILE);
	}

	/* write buf0 to <mbr.bin> file */
	if ((fwrite(buf0, 2, 256, mbr_file) != 256)) {
	    printf("\nError write <mbr.bin> file!\n");
	    exit(EXIT_FAILURE);
	}
	printf("Done\n");
	return 0;
    }
/*-----------------------select mr cmd option----------------------------*/
    else if (!strnicmp(argv[1], "mr", 2)) { 
	if (!(mbr_file = fopen("mbr.bin","rb"))) {
	    printf("<mbr.bin> file was not found!\n");
	    exit (ENFILE);
	}

	printf("Write MBR from <mbr.bin> to disk %s ... ", argv[2]);
	/* read <mbr.bin> file to buf0 */
	if ((fread(buf0, 2, 256, mbr_file) != 256)) {
	    printf("\nError read <mbr.bin> file!\n");
	    exit(EXIT_FAILURE);
	}

	/* write buf0 to disk */
	diskerr = _bios_disk(_DISK_WRITE, &diskreq);
	if (diskerr & 0xff00) {
	    printf("\nError write sector %02d, code 0x%04x\n",
		diskreq.sector, diskerr);
	    exit(EXIT_FAILURE);
	}
	printf("Done\n");
	return 0;
    }
/*-----------------------select ci cmd option----------------------------*/
   /* check that we have both <cfmbr.bin> and <int13r.bin> files */ 
   else if (!strnicmp(argv[1], "ci", 2)) { 
	if (!(cfmbr_file = fopen("cfmbr.bin","rb"))) {
	    printf("<cfmbr.bin> file was not found!\n");
	    exit (ENFILE);
	}
	if (!(int13r_file = fopen("int13r.bin","rb"))) {
	    printf("<int13r.bin> file was not found!\n");
	    exit (ENFILE);
	}

    /* read <cfmbr.bin> file to buf1 */
	if ((fread(buf1, 2, 256, cfmbr_file) != 256)) {
	    printf("Error read <cfmbr.bin> file!\n");
	    exit(EXIT_FAILURE);
	}

    /* read <int13r.bin> file to buf2 */
	if ((fread(buf2, 2, 512, int13r_file) != 512)) {
	    printf("Error read <int13r.bin> file!\n");
	    exit(EXIT_FAILURE);
	}

	/* read original mbr to buf0 */
	printf("Read disk %s MBR to buffer ... ", argv[2]);
	diskerr = _bios_disk(_DISK_READ, &diskreq);
	if (diskerr & 0xff00) {
	    printf("\nError read sector %02d, code 0x%04x\n",
			diskreq.sector, diskerr);
	    exit(EXIT_FAILURE);
	}
	printf("Done\n");

	/* copy 0x01b0 bytes (MBR code) from buf1 to buf0 */
	if (!memcpy(buf0, buf1, 0x01b0)) {
	    printf("Error copy memory buffers!\n");
	    exit(EXIT_FAILURE);
	}

	/* write new INT13h_read code */
	diskreq.sector = 0x02;
	diskreq.buffer = buf2;
	diskreq.nsectors = 0x02;

	printf("Write new int13h code to disk %s ... ", argv[2]);
	diskerr = _bios_disk(_DISK_WRITE, &diskreq);
	if (diskerr & 0xff00) {
	    printf("\nError write sector %02d, code 0x%04x\n",
		diskreq.sector, diskerr);
	    exit(EXIT_FAILURE);
	}
	printf("Done\n");

	/* write new MBR */
	diskreq.sector = 0x01;
	diskreq.buffer = buf0;
	diskreq.nsectors = 0x01;

	printf("Write CF MBR to disk %s ... ", argv[2]);
	diskerr = _bios_disk(_DISK_WRITE, &diskreq);
	if (diskerr & 0xff00) {
	    printf("\nError write sector %02d, code 0x%04x\n",
		diskreq.sector, diskerr);
	    exit(EXIT_FAILURE);
	}
	printf("Done\n");
	return 0;
    }
    printf("Invalid CMD!\n");
return EINVAL;
}