; CFMBR V1.1
;C/H/S
;0/0/1	new MBR and CF loader
;0/0/2	new int13 handler
;0/0/3	new int13 handler

BIOS_SEG_V      EQU     000h
BIOS_OFF_V      EQU     04Ch
MBR_MEM_ADDR	EQU	600h

JMPFAR  MACRO   seg, off	    	;  dynamic JMP FAR SEG:OFF
	db	0eah			;; jmp instruction
        dw      offset	off     	;; offset word
	dw	seg			;; segment word
	endm

CODE	SEGMENT	PARA PUBLIC USE16 'CODE'
MBR	PROC NEAR
ASSUME CS:CODE, DS:NOTHING, ES:NOTHING, SS:NOTHING;
ORG MBR_MEM_ADDR
	cli
	xor	ax, ax
	mov	ss, ax		; SS = 0

	mov	sp, 7C00h
	mov	si, sp

	push	ax
	pop	es		; ES = 0

	push	ax
	pop	ds		; DS = 0
	sti	

	cld	
	mov	di, 600h
	mov	cx, 1b0h

	; copy 1b0 bytes block from 0:07c00h to 0:600h
	rep	movsw

        JMPFAR 0, cf_patch
; 
; mem int 13h seg == ((40h:13h)-1) * 2^6	2^6 ==(2^10 / 2^4)
cf_patch:
	push	es

	mov	si, offset inst_cf_str
	call	print_str
	
        mov     di, word ptr ds:[0413h]
	dec	di			; in DI - mem size (in KB) minus 1

        mov     word ptr ds:[0413h], di	; decrease base mem size by 1KB

	mov	cl, 6
	shl	di, cl		; in DI - mem int13 seg

	mov	si, offset err_cf_str	; prepare err str
	
	mov	es, di
	mov	bx, 0000h	; RAM buffer, base mem - 1KB
        mov     dx, 0080h	; head 0, drive 80h
	mov	cx, 0002h	; read sector 2, cyl 0
	call read_sector
	jnc	@F
err_cf_halt:
	call	print_str
	jmp	short halt_machine

@@:
	add     bx, 512d
	mov	cx, 0003h	; read sector 3, cyl 0
	call read_sector
	jnc	@F
	jmp	short err_cf_halt

@@:
; set vectors
	; mov	ax, BIOS_SEG_V
	; mov	ds, ax	; ds already is equal to zero in the MBR initial code
	mov	bx, BIOS_OFF_V

	push	ds
	lds	bx, ds:[bx]		; ds:bx - BIOS int13h ROM address
	mov	es:[0000h], bx		; ROM BIOS int13h off to mem patch
	mov	es:[0002h], ds		; ROM BIOS int13h seg to mem patch
	pop	ds
					
	cli
	mov	ds:word ptr[BIOS_OFF_V],   0004h	; set new off int13h vector
	mov	ds:word ptr[BIOS_OFF_V+2], es		; set new seg int13h vector
	sti

	mov	si, offset done_cf_str
	call	print_str

	pop	es

; 
movestart:
	mov	si, 7BEh	; addr of first part boot flag
	mov	bl, 4		; number of partitions

@@:
	cmp	byte ptr [si], 80h
	je	prep_boot
	cmp	byte ptr [si], 0
	jnz	inv_pt		; invalid part, if flag non-equal 80h or 00h
	add	si, 10h		; next part descritor
	dec	bl
	jnz	@B
	int	18h		; TRANSFER TO ROM BASIC if no boot part

prep_boot:
	mov	dx, [si]	; dl and dh for int13h
	mov	cx, [si+2]	; cyl and sec for int13h
	mov	bp, si

@@:				; chk for only one part is actived
	add	si, 10h
	dec	bl
	jz	try_boot
	cmp	byte ptr [si], 0
	jz	@B

inv_pt:
	mov	si, offset inv_part_tbl
	call	print_str
halt_machine:
	jmp	short halt_machine

;==================================================================
try_boot:
        ; copy sector to end of base DRAM
	mov	bx, 7C00h	; load OS boot sector to 0:07C00h
	call	read_sector
	jnc	@F
	; print <Load OS error>, if we couldn't read OS boot loader
	mov	si, offset err_load_os
	call	print_str
	jmp	short halt_machine

@@:
	mov	si, offset miss_os
	mov	di, 7DFEh
	cmp	word ptr [di], 0AA55h	; chk MBR signature
	je	@F
	; print <Missing OS>, if OS-boot_loader signature was not found
	call	print_str		; print <Missing OS>, if OS-boot_loader signature was not found
	jmp	short halt_machine

@@:
	mov	si, bp
	JMPFAR	0, 07C00h 		; go to OS loader

;==================================================================
read_sector proc near
	mov	di, 5		; number of retries
@@:
	mov	ax, 0201h	; copy only one sector
        push    di
        int	13h
        pop     di
        jnc     read_ok

        xor     ax, ax
	int	13h		; DISK - RESET DISK SYSTEM
	dec	di
        jnz     @B
        stc			; set error flag

read_ok:
	retn
read_sector endp

;==================================================================
print_str proc near
@@:
	lodsb
	cmp	al, 0
	je	@F		; ending zero was found
	push	si
	mov	bx, 7
	mov	ah, 0Eh
	int	10h		; - VIDEO - WRITE CHARACTER AND	ADVANCE	CURSOR (TTY WRITE)
				; AL = character, BH = display page (alpha modes)
				; BL = foreground color	(graphics modes)
	pop	si
	jmp	short @B
@@:
	retn
print_str endp

;==================================================================
inst_cf_str	db 'Install CFMBR V1.1 ... ', 0
done_cf_str	db 'Done', 0dh, 0ah, 0
err_cf_str	db 'Error',0

inv_part_tbl	db 'Invalid partition table',0
err_load_os	db 'Error loading operating system',0
miss_os		db 'Missing operating system',0
ORG	(MBR_MEM_ADDR + 200h)
MBR	endp
CODE	ends
END	MBR	 
