;V1.1
.286C
;1F2h Sector count
;1F3h Sector number
;1F4h Cylinder high (bits 0-1 are bits 8-9 of 10-bit cylinder number)
;1F5h Cylinder low (bits 0-7 of 10-bit cylinder number)
;1F6h Drive & Head
; used
; [bp+3] number of sectors
; [bp+7] D/H reg

;MAX_DRV	EQU 81h

CODE	SEGMENT	PARA PUBLIC 'CODE'
ORG 00h
ASSUME CS:CODE, DS:CODE, ES:CODE, SS:CODE;

START:

bios_addr_off	dw	?
bios_addr_seg	dw	?

new_int13h proc far
	cmp	dl, 80h
	jb	orig_int13h	; floppy, jump to original handler

	cmp	dl, MAX_DRV	; if disk more than MAX_DRV, jump to original handler
	ja	orig_int13h

	cmp	ah, 02h		; chk command, if read command, jump to new handler
	je	new_read

orig_int13h:
	pushf
	cli
	call dword ptr cs:[bios_addr_off]
	retf	2

new_read:		; flg	; 30	22
	sti		; seg	; 28	20
	cld		; off	; 26	18
	push	ax		; 24	16
	push	cx		; 22	14
	push	dx		; 20	12
	push    bx		; 18	10
	push	si		; 16	08
	push	di		; 14	06
	push	ds		; 12	04
	push	es		; 10	02
	push	bp		
	mov	bp, sp		; 00	00
	push	40h
	pop	ds		; DS to	BDA
	assume ds:nothing
	call	read_hdd

	mov	[bp+16], ax
	mov	ax, 201h
	jc	@F
	and	byte ptr [bp+22], 0FEh		; clear flag CF
	dec	ax

@@:
	or	[bp+22], ax			; set (or unset) flag CF and IF
	pop	bp
	pop	es
	pop	ds
	pop	di
	pop	si
	pop	bx
	pop	dx
	pop	cx
	pop	ax
	iret
new_int13h	endp


; 
;		S u b r	o u t i	n e
read_hdd proc near
	mov	word ptr [bp+16], 512
	call	chk_data_seg
	jc	ata_read_exit
	mov	ah, 20h

ata_read:
	call	chkdrv_init_dh
	jc	ata_read_exit
	call	wait_ready
	jb	ata_read_exit
	mov	di, bx
	mov	si, es

	test	byte ptr ds:76h, 0C0h
	jz	skip_dis_retry
	inc	ah
skip_dis_retry:
	call	tf_write
	mov	al, ah
	call	ata_cmd

ata_read_sector:
	call	wait_int
	jb	ata_read_exit
	call	chk_stat_err
	jnb	ata_r_no_err
	cmp	byte ptr ds:74h, 11h
	jnz	set_ah_err_code	; if ECC corrected error, then continue
ata_r_no_err:   
	mov	es, si
	mov	cx, 100h
	mov	dx, 1F0h
	cli	
	cld	
	rep ins	    word ptr es:[di], dx
	sti	
	dec	byte ptr cs:[sec_num]
	jnz	ata_read_sector

set_ah_err_code:
	mov	ah, ds:74h
	or	ah, ah
	jz	ata_read_exit
	stc	
ata_read_exit:  
	retn
read_hdd endp
        

; 
;		S u b r	o u t i	n e
chk_data_seg	proc near
	push	dx
	push	ax
	mov	ax, bx
	shr	ax, 4		; buf offset / 16
	mov	dx, es
	add	ax, dx		; (buf_offset/16)+seg to es seg
	mov	es, ax
	and	bx, 0Fh		; select low nibble
	xor	ax, ax		; ax = 0
	mov	dx, 1
	sub	ax, bx		; ax = 0 - bx
	sbb	dx, 0
	div	word ptr [bp+16] ; div ax to 512
	mov	dl, al
	pop	ax
	cmp	dl, al
	pop	dx
	jnb	chk_seg_exit
	mov	ah, 9
	mov	ds:74h,	ah	; seg across 64KB bound
chk_seg_exit:
	retn	
chk_data_seg	endp


; 
;		S u b r	o u t i	n e
chkdrv_init_dh	proc near		; CODE XREF: F:A22Bp F:A2C4p ...
	push	ax
	mov	al, ds:75h	; al = HDD's quantity
	mov	ah, dl
	and	ah, 7Fh		; ah = current hdd number
	cmp	al, ah
	pop	ax
	jbe	chkdrv_err_ex

	push	dx
	push	cx
	mov	cl, 0A0h
	test	dl, 1
	jz	@F
	mov	cl, 0B0h
@@:
	and	dh, 0Fh
	or	cl, dh
	mov	byte ptr cs:[drv_head], cl
	pop	cx
	pop	dx
	clc	
	retn	

chkdrv_err_ex:		
	mov	ah, 1
	mov	ds:74h,	ah
	stc	
	retn	
chkdrv_init_dh	endp


; 
;		S u b r	o u t i	n e
ata_cmd		proc near
		call	wait_no_busy
		cli	
		push	ax
		mov	byte ptr ds:8Eh, 0
		in	al, 0A1h	; Interrupt Controller #2, 8259A
		jmp	short $+2
		jmp	short $+2
		and	al, 0BFh
		out	0A1h, al	; Interrupt Controller #2, 8259A
		in	al, 21h		; Interrupt controller,	8259A.
		jmp	short $+2
		jmp	short $+2
		and	al, 0FBh
		out	21h, al		; Interrupt controller,	8259A.
		jmp	short $+2
		jmp	short $+2
		pop	ax

		push	dx
		mov	dx, 1F7h
		out	dx, al		; AT hard disk
					; command register:
					; 1?H =	Restore	to cylinder 0
					; 7?H =	Seek to	cylinder
					; 2?H =	Read sector
					; 3xH =	Write sector
					; 50H =	Format track
					; 4xH =	verify read
					; 90H =	diagnose
					; 91H =	set parameters for drive
		pop	dx

		sti	
		retn	
ata_cmd		endp


; 
;		S u b r	o u t i	n e
tf_write	proc near
	push	ax
	push	dx
	mov	dx, 1F2h
	mov	byte ptr cs:[sec_num], al
	out	dx, al		; Load sector count ATA register
	jmp	short $+2
	jmp	short $+2

	mov	al, cl
	and	al, 3Fh		; mask sector number
	mov	dx, 1F3h
	out	dx, al		; Load sector number ATA register
	jmp	short $+2
	jmp	short $+2

	mov	al, ch
	mov	dx, 1F4h
	out	dx, al		; Load cylinder low ATA register
	jmp	short $+2
	jmp	short $+2

	mov	al, cl
	shr	al, 6		; in AL 2 high bit of cyl number

if 0
	mov	dl, [bp+13]	; in dl - [bp+21], it is dh - head
	shr	dl, 4		; in dl - head / 16
	and	dl, 0Ch		; select bits 2 and 3
	or	al, dl		; OR bits 6 and 7 of head with high cyl ??
endif
	mov	dx, 1F5h
	out	dx, al		; Load cylinder high ATA register
	jmp	short $+2
	jmp	short $+2

	mov	dx, 1F6h
	mov	al, byte ptr cs:[drv_head]
	out	dx, al		; AT hard disk controller:
				; Drive	& Head.
				; Read/Write: bits indicate head, drive	for operation
	pop	dx
	pop	ax
	retn	
tf_write	endp


; 
;		S u b r	o u t i	n e
wait_ready  proc near
	push	dx
	push	cx
	push	ax
	mov	cx, 3072

wait_ready_loop:
	call	wait_no_busy		; read status reg with wait for ready
        jc      wait_ready_err_exit

	mov	dx, 1F6h
	mov	al, byte ptr cs:[drv_head]
	out	dx, al

	push	cx
	mov	cx, 2
	call	dly_cx
	pop	cx
	mov	dx, 1F7h
        in      al, dx
				; status register bits:
				; 0: 1=prev cmd	error
				; 2: Corrected data
				; 3: Data Request. Buffer is busy
				; 4: Seek completed
				; 5: Write fault
				; 6: Drive ready (unless bit 4=0)
				; 7: Busy

        mov     ah, 0AAh
	test	al, 40h
        jz      @F
        mov     ah, 40h
	test	al, 10h
        jnz     wait_ready_ok_exit	; wait DRQ+DRSC

@@:
        loop    wait_ready_loop
	mov	ds:8Ch,	al
	mov	ds:74h,	ah

wait_ready_err_exit:
	pop	cx
	mov	al, cl
	pop	cx
	pop	dx
	stc	
	retn	

wait_ready_ok_exit:
	pop	ax
	pop	cx
	pop	dx
	clc	
	retn	
wait_ready endp


; 
;		S u b r	o u t i	n e
; read status reg with wait for ready
wait_no_busy	proc near
	sti	
	push	cx
	push	dx
	push	ax

	mov	cx, 10h
	mov	dx, 1F7h
	mov	ah, 80h

@@:
	push	cx
	xor	cx, cx
	call    dly_cx_ah
	pop	cx
	jnc	wnb_ex2
	loop	@B

	in	al, dx
	mov	ds:8Ch,	al

	pop	ax
	mov	ah, 80h
	mov	ds:74h,	ah
	jmp	short wnb_ex1

wnb_ex2:
	pop	ax
wnb_ex1:
	pop	dx
	pop	cx
	retn	
wait_no_busy	endp


; 
;		S u b r	o u t i	n e
dly_cx_ah	proc near
	push	ax

dly_cx_ah_loop:
	in	al, dx
	test	ah, al
	jz	dly_cx_ah_exit

@@:				
	jmp	short $+2
	in	al, 61h
	test	al, 10h
	jz	@B		; wait 1

	dec	cx
	jz	dly_cx_ah_err
	in	al, dx
	test	ah, al
	jz	dly_cx_ah_exit

@@:
	jmp	short $+2
	in	al, 61h
	test	al, 10h
	jnz	@B		; wait 0

	dec	cx
	jnz	dly_cx_ah_loop
dly_cx_ah_err:		
	stc	
dly_cx_ah_exit:
	pop	ax
	retn	
dly_cx_ah	endp


; 
;		S u b r	o u t i	n e
wait_int	proc near
	clc	
	mov	ax, 9000h
	int	15h		; OS HOOK - DEVICE BUSY	LOOP (AT,XT2,XT286,CONV,PS)
				; AL = disk
	sti	
	jc	w_int_err_exit
	mov	bx, 8Eh
	mov	dx, 10h

wait_int_loop:
	xor	cx, cx
        call    dly_w_int
	jnc	w_int_ok_exit
	dec	dx
	jnz	wait_int_loop

w_int_err_exit:
	mov	ah, 80h
	mov	ds:74h,	ah
	jmp	short w_int_ret
; 
w_int_ok_exit:
	mov	byte ptr ds:8Eh, 0

w_int_ret:
	retn	
wait_int endp

                
; 
;		S u b r	o u t i	n e
chk_stat_err    proc near
	push	dx
	push	ax
	mov	dx, 1F7h
	in	al, dx		; AT hard disk
				; status register bits:
				; 0: 1=prev cmd	error
				; 2: Corrected data
				; 3: Data Request. Buffer is busy
				; 4: Seek completed
				; 5: Write fault
				; 6: Drive ready (unless bit 4=0)
				; 7: Busy
	mov	ds:8Ch,	al

	mov	ah, 11h
	test	al, 4
	jnz	chk_es_ex

	mov	ah, 0CCh
	test	al, 20h
	jnz	chk_es_ex

	mov	ah, 0AAh
	test	al, 40h
	jz	chk_es_ex

	mov	ah, 40h
	test	al, 10h
	jz	chk_es_ex

	mov	ah, 0
	test	al, 1
	jz	chk_es_ex

	mov	dx, 1F1h
	in	al, dx		; AT hard disk controller
					; Error	register. Bits for last	error:
					; 0: Data Address Mark not found
					; 1: Track 0 Error
					; 2: Command aborted
					; 4: Sector ID not found
					; 6: ECC Error:	Uncorrectable data error
					; 7: Bad block
	mov	ds:8Dh,	al

	mov	ah, 2
	test	al, 1
	jnz	chk_es_ex

	mov	ah, 1
	test	al, 4
	jnz	chk_es_ex

	mov	ah, 4
	test	al, 10h
	jnz	chk_es_ex

	mov	ah, 10h
	test	al, 40h
	jnz	chk_es_ex

	mov	ah, 40h
	test	al, 2
	jnz	chk_es_ex

	mov	ah, 0Ah
	test	al, 80h
	jnz	chk_es_ex

	mov	ah, 0BBh
chk_es_ex:
	mov	ds:74h,	ah
	pop	dx
	mov	al, dl
	pop	dx
	or	ah, ah
	jz      @F
	stc	
@@: 
	retn	
chk_stat_err	endp


; 
;		S u b r	o u t i	n e
dly_w_int		proc near
	push	ax

dly_w_int_l1:
	test	byte ptr [bx], 80h
	jnz	dly_w_int_ok_ex		

@@:
	jmp	short $+2
	in	al, 61h
	test	al, 10h
	jz	@B		; wait 1
	dec	cx
	jz	dly_w_int_err_ex
	test	byte ptr [bx], 80h
	jnz	dly_w_int_ok_ex

dly_w_int_l2:
	jmp	short $+2
	in	al, 61h
	test	al, 10h
	jnz	dly_w_int_l2	; wait 0
	dec	cx
	jnz	dly_w_int_l1
dly_w_int_err_ex:
	stc	
dly_w_int_ok_ex:
	pop	ax
	retn	
dly_w_int		endp


; 
;		S u b r	o u t i	n e
dly_cx		proc near
	push	ax
	add	cx, 2

dly_cx_l1:
	jmp	short $+2
	in	al, 61h
	test	al, 10h		; test mem refr signal
	jz	dly_cx_l1	; wait 1
	dec	cx
	jz	dly_cx_end

dly_cx_l2:
	jmp	short $+2
	in	al, 61h
	test	al, 10h		; test mem refr signal
	jnz	dly_cx_l2	; wait 0
	dec	cx
	jnz	dly_cx_l1

dly_cx_end:
	pop	ax
	retn	
dly_cx		endp

sec_num		db ?		; [bp+3] number of sectors
drv_head	db ?		; [bp+7] D/H reg

db "INT13R V1.1",0	; just label for this bin file
db "MAX_DRV = ", ((MAX_DRV AND 0F0h) SHR 4) OR 30h, (MAX_DRV AND 0Fh) OR 30h, 'h'
ORG	1024d

CODE	ENDS
END	START
